import dataclasses
import logging
from dataclasses import dataclass

import utensil
from utensil import constant
from utensil.general.logger import parse_log_level

logger = utensil.get_logger(__name__)


DEFAULT_CONNECTOR_VERBOSE = logging.INFO


@dataclass
class ConnectorInformation:
    ip: str
    port: int
    user: str
    pds: str = dataclasses.field(repr=False)  # to protect password accidentally shown in log
    database: str


@dataclass
class ConnectorPoolInformation:
    connector_info: ConnectorInformation
    pool_size: int
    verbose: int


def get_connector_pool_info(tag_name: str):
    TAG_PREFIX = 'CONNECTION_INFO_OF_'
    trial_tags = []
    if tag_name.startswith(TAG_PREFIX):
        trial_tags.append(tag_name)
    trial_tags.append(TAG_PREFIX + tag_name)

    for trial in trial_tags:
        if constant.config.has_section(trial):
            connection_info_config = constant.config[trial]
            return ConnectorPoolInformation(
                connector_info=ConnectorInformation(
                    ip=connection_info_config.get('Ip'),
                    port=int(connection_info_config.get('Port')),
                    user=connection_info_config.get('User'),
                    pds=connection_info_config.get('Password'),
                    database=connection_info_config.get('ServiceName')
                ),
                pool_size=int(connection_info_config.get('PoolSize', 1)),
                verbose=parse_log_level(connection_info_config.get('Verbose', DEFAULT_CONNECTOR_VERBOSE))
            )
    err_str = '" and "'.join(trial_tags)
    raise ValueError(f'Cannot find connection info match "{err_str}".')





def chunks(iterable, n):
    """
    Yield successive n-sized chunks from lst.
    """
    if n == 0:
        yield from []
    elif isinstance(iterable, (list, str, tuple)):
        for i in range(0, len(iterable), n):
            yield iterable[i:i + n]
    else:
        ret = []
        for it in iterable:
            ret.append(it)
            if len(ret) == n:
                yield ret
                ret = []
        if len(ret) > 0:
            yield ret



import logging
import os
import socket
from dataclasses import dataclass
from logging import handlers
from typing import Union, Iterable

from utensil import constant


def parse_log_level(level):
    if isinstance(level, str):
        if level.upper() == 'NOTSET':
            return logging.NOTSET
        elif level.upper() == 'DEBUG':
            return logging.DEBUG
        elif level.upper() == 'INFO':
            return logging.INFO
        elif level.upper() == 'WARNING':
            return logging.WARNING
        elif level.upper() == 'WARN':
            return logging.WARN
        elif level.upper() == 'ERROR':
            return logging.ERROR
        elif level.upper() == 'FATAL':
            return logging.FATAL
        elif level.upper() == 'CRITICAL':
            return logging.CRITICAL
    try:
        return int(level)
    except ValueError as e:
        raise ValueError(f"invalid log level: '{level}'") from e


@dataclass
class LoggerConfig:
    level: Union[str, int] = constant.LOG.get('Level')
    handlers: Iterable[logging.Handler] = None,
    format: str = '{asctime:s}.{msecs:06.0f} ' + constant.HOST_INFO.get('HostName') + ' ' + socket.gethostname() + \
                  ' {processName:s}({process:d}) {threadName:s}({thread:d}) {levelname:s} ' \
                  '({name:s}.{funcName:s}) {message:s} '
    style: str = '{'
    datefmt: str = '%Y-%m-%d %I:%M:%S'

    def __post_init__(self):

        _logger_handlers = []
        if constant.LOG.get('Stream', 'NOTSET') != 'NOTSET':
            handler = logging.StreamHandler()
            handler.setLevel(parse_log_level(constant.LOG.get('Stream', 'NOTSET')))
            _logger_handlers.append(handler)
        if constant.LOG.get('Syslog', 'NOTSET') != 'NOTSET':
            handler = handlers.SysLogHandler()
            handler.setLevel(parse_log_level(constant.LOG.get('Syslog', 'NOTSET')))
            _logger_handlers.append(handler)
        if constant.LOG.get('File', 'NOTSET') != 'NOTSET':
            log_file_name = constant.LOG.get('FilePrefix', 'log')
            if not os.path.isdir(os.path.dirname(log_file_name)):
                os.makedirs(os.path.dirname(log_file_name))
            handler = handlers.WatchedFileHandler(log_file_name)
            handler.setLevel(parse_log_level(constant.LOG.get('File', 'NOTSET')))
            _logger_handlers.append(handler)

        self.level = parse_log_level(self.level)
        if self.handlers == (None,):
            self.handlers = _logger_handlers


def get_logger(name, logger_config=None):
    if logger_config is None:
        logger_config = LoggerConfig()
    logger = logging.getLogger(name)
    logger.setLevel(logger_config.level)
    formatter = logging.Formatter(
        fmt=logger_config.format, datefmt=logger_config.datefmt, style=logger_config.style
    )
    for handler in logger_config.handlers:
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


class _DummyLogger:

    def setLevel(self, level):
        pass

    def debug(self, msg, *args, **kwargs):
        pass

    def info(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def warn(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass

    def exception(self, msg, *args, exc_info=True, **kwargs):
        pass

    def critical(self, msg, *args, **kwargs):
        pass

    fatal = critical

    def log(self, level, msg, *args, **kwargs):
        pass

    def findCaller(self, stack_info=False, stacklevel=1):
        pass

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        pass

    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
        pass

    def handle(self, record):
        pass

    def addHandler(self, hdlr):
        pass

    def removeHandler(self, hdlr):
        pass

    def hasHandlers(self):
        pass

    def callHandlers(self, record):
        pass

    def getEffectiveLevel(self):
        pass

    def isEnabledFor(self, level):
        pass

    def getChild(self, suffix):
        pass


DUMMY_LOGGER = _DummyLogger()



from ._random_search import (
    RandomizedParam,
    BooleanParam,
    UniformBetweenParam,
    ExponentialBetweenParam,
    RandomizedChoices,
    RandomizedConfig,
    RandomizedDispatcher,
    SeededConfig,
    RandomSearch,
    ModelScore
)



import os
from configparser import ConfigParser, ExtendedInterpolation

from utensil.constant._constant import *

__all__ = ['PROJECT', 'HOST_INFO', 'LOG']

if 'UTENSIL_PROJECT_ROOT' in os.environ:
    PROJECT_ROOT = os.environ['UTENSIL_PROJECT_ROOT']
else:
    PROJECT_ROOT = '.'

if 'UTENSIL_CONFIG' in os.environ:
    UTENSIL_CONFIG = os.environ['UTENSIL_CONFIG']
else:
    UTENSIL_CONFIG = 'utensil.ini'


config = ConfigParser(interpolation=ExtendedInterpolation())

config.read_string(f"""
[PROJECT]
ProjectRoot = {PROJECT_ROOT}
ConfigPath = {PROJECT_ROOT}/{UTENSIL_CONFIG}
ProjectName = .utensil
ProjectAbbr = ${{ProjectName}}
ProjectState = dev

[HOST_INFO]
HostName = localhost

[LOG]
Dir = {PROJECT_ROOT}/${{PROJECT:ProjectAbbr}}/log
Stream = info
Syslog = notset
File = info
FilePrefix = ${{Dir}}/${{PROJECT:ProjectAbbr}}.log
Level = info
MaxMessageLen = 60000
""")

if os.path.isfile(os.path.normpath(config['PROJECT'].get('ConfigPath'))):
    config.read_file(open(os.path.normpath(config['PROJECT'].get('ConfigPath'))))

config['PROJECT']['ProjectRoot'] = os.path.normpath(config['PROJECT']['ProjectRoot'])
config['PROJECT']['ProjectName'] = os.path.normpath(config['PROJECT']['ProjectName'])
config['PROJECT']['ProjectAbbr'] = os.path.normpath(config['PROJECT']['ProjectAbbr'])
config['PROJECT']['ConfigPath'] = os.path.normpath(config['PROJECT']['ConfigPath'])

config['LOG']['Dir'] = os.path.normpath(config['LOG']['Dir'])
config['LOG']['FilePrefix'] = os.path.normpath(config['LOG']['FilePrefix'])

PROJECT = config['PROJECT']
HOST_INFO = config['HOST_INFO']
LOG = config['LOG']

del os, ConfigParser, ExtendedInterpolation
del PROJECT_ROOT
del UTENSIL_CONFIG



import unittest as ut
from importlib import reload
import os


class TestDefaultConstant(ut.TestCase):

    def setUp(self) -> None:
        self.project_root = os.path.dirname(os.path.dirname(__file__))
        try:
            self.original_project_root = os.environ['UTENSIL_PROJECT_ROOT']
        except KeyError:
            pass
        os.environ['UTENSIL_PROJECT_ROOT'] = self.project_root

    def tearDown(self) -> None:
        try:
            os.environ['UTENSIL_PROJECT_ROOT'] = self.original_project_root
        except AttributeError:
            del os.environ['UTENSIL_PROJECT_ROOT']

    def test_PROJECT_ProjectRoot(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(self.project_root, constant.PROJECT['ProjectRoot'])

    def test_PROJECT_ProjectName(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('.utensil', constant.PROJECT['ProjectName'])

    def test_PROJECT_ProjectAbbr(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('.utensil', constant.PROJECT['ProjectAbbr'])

    def test_PROJECT_ProjectState(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('dev', constant.PROJECT['ProjectState'])

    def test_PROJECT_ConfigPath(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join(self.project_root, 'utensil.ini'), constant.PROJECT['ConfigPath'])

    def test_HOST_INFO_HostName(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('localhost', constant.HOST_INFO['HostName'])

    def test_LOG_Dir(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join(self.project_root, '.utensil', 'log'), constant.LOG['Dir'])

    def test_LOG_Stream(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('info', constant.LOG['Stream'])

    def test_LOG_Syslog(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('notset', constant.LOG['Syslog'])

    def test_LOG_File(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('info', constant.LOG['File'])

    def test_LOG_FilePrefix(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join(self.project_root, '.utensil', 'log', '.utensil.log'), constant.LOG['FilePrefix'])

    def test_LOG_Level(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('info', constant.LOG['Level'])

    def test_LOG_MaxMessageLen(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('60000', constant.LOG['MaxMessageLen'])


class TestModifiedConstant(ut.TestCase):
    def setUp(self) -> None:
        self.project_root = os.path.dirname(__file__)
        try:
            self.original_project_root = os.environ['UTENSIL_PROJECT_ROOT']
        except KeyError:
            pass
        try:
            self.original_utensil_config = os.environ['UTENSIL_CONFIG']
        except KeyError:
            pass
        self.utensil_config = 'utensil_test1.ini'
        os.environ['UTENSIL_CONFIG'] = self.utensil_config
        os.environ['UTENSIL_PROJECT_ROOT'] = self.project_root

    def tearDown(self) -> None:
        try:
            os.environ['UTENSIL_PROJECT_ROOT'] = self.original_project_root
        except AttributeError:
            del os.environ['UTENSIL_PROJECT_ROOT']
        try:
            os.environ['UTENSIL_CONFIG'] = self.original_utensil_config
        except AttributeError:
            del os.environ['UTENSIL_CONFIG']

    def test_PROJECT_ProjectRoot(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(self.project_root, constant.PROJECT['ProjectRoot'])

    def test_PROJECT_ProjectName(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('proj_name', constant.PROJECT['ProjectName'])

    def test_PROJECT_ProjectAbbr(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('proj_name_proj_abbr', constant.PROJECT['ProjectAbbr'])

    def test_PROJECT_ProjectState(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('proj_state', constant.PROJECT['ProjectState'])

    def test_PROJECT_ConfigPath(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join(self.project_root, self.utensil_config), constant.PROJECT['ConfigPath'])

    def test_HOST_INFO_HostName(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('my_host', constant.HOST_INFO['HostName'])

    def test_LOG_Dir(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join('proj_name_proj_abbr', 'log_test'), constant.LOG['Dir'])

    def test_LOG_Stream(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('notset', constant.LOG['Stream'])

    def test_LOG_Syslog(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('info', constant.LOG['Syslog'])

    def test_LOG_File(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('notset', constant.LOG['File'])

    def test_LOG_FilePrefix(self):
        from utensil import constant
        reload(constant)
        self.assertEqual(os.path.join('proj_name_proj_abbr', 'log_test', 'proj_name.log123'), constant.LOG['FilePrefix'])

    def test_LOG_Level(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('debug', constant.LOG['Level'])

    def test_LOG_MaxMessageLen(self):
        from utensil import constant
        reload(constant)
        self.assertEqual('12345', constant.LOG['MaxMessageLen'])


if __name__ == '__main__':
    ut.main()



from __future__ import annotations

import abc
import hashlib
import os
from collections import namedtuple
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from typing import Any, Tuple, Dict, Union, TypeVar

from utensil.general.logger import DUMMY_LOGGER

try:
    import numpy as np
except ImportError as e:
    raise e

try:
    import pandas as pd
except ImportError as e:
    raise e


class RandomizedParam(abc.ABC):
    @abc.abstractmethod
    def from_random(self, r):
        raise NotImplemented


@dataclass
class BooleanParam(RandomizedParam):
    def from_random(self, r):
        return r > 0.5


@dataclass
class UniformBetweenParam(RandomizedParam):
    left: Any
    right: Any
    otype: type

    def from_random(self, r):
        return self.otype(r * (self.right - self.left) + self.left)


@dataclass
class ExponentialBetweenParam(RandomizedParam):
    left: Any
    right: Any
    otype: type

    def from_random(self, r):
        log_right = np.log(self.right)
        log_left = np.log(self.left)
        return self.otype(np.exp(r * (log_right - log_left) + log_left))


@dataclass(init=False)
class RandomizedChoices(RandomizedParam):
    choice: Tuple[Enum]

    def __init__(self, *args: Enum):
        self.choice = args

    def from_random(self, r):
        nr_choices = len(self.choice)
        return [c for c in self.choice][-1 if r == 1 else int(r * nr_choices)]


@dataclass
class RandomizedDispatcher:
    key_names: Union[str, Tuple[str]]
    dispatch: Dict[Any, RandomizedConfig]


TRandomizedConfig = TypeVar('TRandomizedConfig', bound='RandomizedConfig')


class RandomizedConfig(abc.ABC):
    def get_config(self, model_id, seed=0) -> Tuple[Dict[str, Any], TRandomizedConfig]:
        kwargs = {}
        dispatchers = {}
        params = {}
        model_r = {}
        for k, v in vars(self).items():
            if isinstance(v, RandomizedParam):
                params[k] = v
            elif isinstance(v, RandomizedDispatcher):
                dispatchers[k] = v
                kwargs[k] = v
            else:
                kwargs[k] = v
                model_r[k] = None

        base = 2 ** int(np.log2(model_id + 1))
        offset = model_id + 1 - base
        sd = int.from_bytes(hashlib.sha256(str(seed + base).encode()).digest()[:4], 'big')
        rng = np.random.default_rng(sd)
        linspace = np.linspace(0, 1, base + 1)
        rand_space = rng.random(size=(base, len(params))) * (linspace[1] - linspace[0]) + linspace[:-1].reshape(-1, 1)

        for i in range(len(params)):
            rng.shuffle(rand_space[:, i])

        for (k, v), r in zip(params.items(), rand_space[offset]):
            model_r[k] = r
            kwargs[k] = v.from_random(r)

        model_c = self.__class__(**kwargs)
        for var_name, dispatcher in dispatchers.items():
            if isinstance(dispatcher.key_names, str):
                key = vars(model_c)[dispatcher.key_names]
            else:
                key = tuple(vars(model_c)[kn] for kn in dispatcher.key_names)
            r, vars(model_c)[var_name] = dispatcher.dispatch[key].get_config(model_id, seed=seed)
            model_r[var_name] = r
        return model_r, model_c

    def to_dict(self) -> Dict[str, Any]:
        d = {}
        for k, v in vars(self).items():
            if isinstance(v, RandomizedConfig):
                d[k] = v.to_dict()
            else:
                d[k] = v
        return d

    def to_plain_dict(self, sep=':') -> Dict[str, Any]:
        d = {}
        for k, v in vars(self).items():
            if isinstance(v, RandomizedConfig):
                for vk, vv in v.to_plain_dict().items():
                    d[f'{k}{sep}{vk}'] = vv
            else:
                d[k] = v
        return d


@dataclass
class SeededConfig:
    cid: int
    base_seed: int
    seed_r: Dict[str, Any]
    config: TRandomizedConfig
    config_temp: TRandomizedConfig

    @classmethod
    def from_config_template(cls, config_temp: RandomizedConfig, model_id: int, seed: int):
        seed_r, config = config_temp.get_config(model_id=model_id, seed=seed)
        return cls(cid=model_id, base_seed=seed, seed_r=seed_r, config=config, config_temp=config_temp)


ModelScore = namedtuple('ModelScore', ['model', 'score'])


class RandomSearch(abc.ABC):
    def __init__(self, logger=None):
        self.logger = DUMMY_LOGGER if logger is None else logger

    @abc.abstractmethod
    def get_xy(self, tr_path, te_path) -> Tuple[pd.DataFrame, pd.DataFrame]:
        raise NotImplementedError

    @abc.abstractmethod
    def model_scores_to_csv(self, model_scores) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def do_training(self, sd_config: SeededConfig, train_x, train_y, idx) -> ModelScore:
        raise NotImplementedError

    def train(self, tr_path, te_path, config_temp, model_id_range=None, seed=0):
        if model_id_range is None:
            model_id_range = range(10)

        x, y = self.get_xy(tr_path, te_path)

        te_x = x[y.isna()]
        train_x = x[~y.isna()]
        train_y = y[~y.isna()]

        rng = np.random.default_rng(seed)
        idx = np.arange(train_x.shape[0])
        rng.shuffle(idx)

        if not os.path.exists('submit'):
            os.mkdir('submit')
        model_scores = {}
        best_model = None

        for mid in model_id_range:
            self.logger.info(f'model_id={mid}: initialize')

            sd_config = SeededConfig.from_config_template(config_temp=config_temp, model_id=mid, seed=seed)
            model, score = None, None
            try:
                model, score = self.do_training(sd_config, train_x, train_y, idx)
            except Exception:
                self.logger.warning(f'model_id={sd_config.cid}: invalid config for model_id={sd_config.cid}, {seed=}',
                               stack_info=True)

            self.logger.info(f'model_id={sd_config.cid}: {score=}')

            # record model_id, model_config and score
            model_scores[sd_config.cid] = sd_config.config.to_plain_dict()
            assert 'score' not in model_scores[sd_config.cid]
            model_scores[sd_config.cid]['score'] = score
            self.model_scores_to_csv(model_scores)

            # keep the best model
            if score is not None and (best_model is None or best_model[0] < score):
                best_model = (score, deepcopy(model), sd_config)

                # use the current best model to generate csv for submission
                te_xpy = np.empty(shape=(te_x.shape[0], 2), dtype=int)
                te_xpy[:, 0] = np.arange(te_x.shape[0]) + 1
                te_xpy[:, 1] = best_model[1].predict(te_x)
                submit_path = os.path.join('submit', f'{__name__}_{int(np.round(score * 1e5))}_{sd_config.cid:04d}.csv')
                pd.DataFrame(te_xpy, columns=['ImageId', 'Label']).to_csv(submit_path, index=False)
            if best_model is not None:
                self.logger.info(
                    f'model_id={sd_config.cid}: current best model_id={best_model[2].cid}, score={best_model[0]}, '
                    f'config={best_model[2].config.to_plain_dict()}')

        return best_model, model_scores
